import numpy as np


def query_func_meta():
    return _dict


def Wingweight_failmodel(x, theta, p):
    """Given x and theta, return matrix of [row x] times [row theta] of values."""
    ind = np.argmin(np.abs(_dict['fail_array'] - p))
    if p < 0.05:
        ind -= 10
    c = _dict['c_array'][ind]
    f = Wingweight_model(x, theta)
    wheretoobig = np.where((f / Wingweight_true(x)) > c)
    f[wheretoobig[0], wheretoobig[1]] = np.nan
    return f


def Wingweight_failmodel_random(x, theta, p):
    f = Wingweight_model(x, theta)
    wheretoobig = np.where(
        np.random.choice([0, 1], f.shape, replace=True, p=[1 - p, p]))
    f[wheretoobig[0], wheretoobig[1]] = np.nan

    return f


# def Wingweight_failmodel_MAR(x, theta, p):
#     """Given x and theta, return matrix of [row x] times [row theta] of values."""
#     f = Wingweight_model(x, theta)
#     mask = MAR_mask(f, p=p, p_obs=0.5)
#     f[mask] = np.nan
#     return f


def Wingweight_model(x, theta):
    """Given x and theta, return matrix of [row x] times [row theta] of values."""

    theta = tstd2theta(theta)
    x = xstd2x(x)
    p = x.shape[0]
    n = theta.shape[0]

    theta_stacked = np.repeat(theta, repeats=p, axis=0)
    x_stacked = np.tile(x.astype(float), (n, 1))

    f = Wingweight_vec(x_stacked, theta_stacked).reshape((n, p))
    return f.T


def Wingweight_true(x):
    """Given x, return matrix of [row x] times 1 of values."""
    # assume true theta is [0.5]^d
    theta0 = np.atleast_2d(np.array([0.5] * _dict['thetadim']))
    f0 = Wingweight_model(x, theta0)

    return f0


def Wingweight_vec(x, theta):
    (Sw, Wfw, tc, Nz, Wdg, Wp) = np.split(x, x.shape[1], axis=1)
    (A, LamCaps, q, lam) = np.split(theta, theta.shape[1], axis=1)

    LamCaps *= np.pi / 180.

    fact1 = 0.036 * Sw ** 0.758 * Wfw ** 0.0035
    fact2 = (A / (np.cos(LamCaps) ** 2)) ** 0.6
    fact3 = q ** 0.006 * lam ** 0.04
    fact4 = (100 * tc / np.cos(LamCaps)) ** (-0.3)
    fact5 = (Nz * Wdg) ** 0.49

    term1 = Sw * Wp

    W = (fact1 * fact2 * fact3 * fact4 * fact5 + term1).reshape(-1)
    return W


def xstd2x(xstd):
    if xstd.ndim < 1.5:
        xstd = xstd[:, None].T
    # (Sws, Wfws, tcs, Nzs, Wdgs, Wps) = np.split(xstd, xstd.shape[1], axis=1)

    bounds = np.array([[150, 200],
                       [220, 300],
                       [0.08, 0.18],
                       [2.5, 6],
                       [1700, 2500],
                       [0.025, 0.08]])

    x = bounds[:, 0] + xstd * (bounds[:, 1] - bounds[:, 0])
    return x


def tstd2theta(tstd):
    """Given standardized theta in [0, 1]^d, return non-standardized theta."""
    if tstd.ndim < 1.5:
        tstd = tstd[:, None].T
    # (A, LamCaps, q, lam) = np.split(tstd, tstd.shape[1], axis=1)

    bounds = np.array([[6, 10],
                       [-10, 10],
                       [16, 45],
                       [0.5, 1]])

    theta = bounds[:, 0] + tstd * (bounds[:, 1] - bounds[:, 0])
    return theta


def testwingweight():
    np.seterr(all='raise')
    import matplotlib.pyplot as plt
    import scipy.stats.qmc as qmc
    sampler = qmc.LatinHypercube(d=_dict['thetadim'])
    cs = np.arange(0.5, 1.5, 0.0005)
    failss = np.zeros((2, *cs.shape))
    for k in np.arange(2):
        x = np.random.uniform(size=(15, _dict['xdim']))
        theta = sampler.random(1000)
        fails = np.zeros_like(cs)
        for i, c in enumerate(cs):
            f = Wingweight_failmodel(x, theta, c)
            fails[i] = np.isnan(f).mean()
        failss[k] = fails.copy()
    plt.plot(cs, failss.T)
    plt.show()
    np.seterr()
    return cs, failss


_dict = {
    'function': 'Wingweight',
    'xdim': 6,
    'thetadim': 4,
    'xbounds': [[150, 200],
                [220, 300],
                [0.08, 0.18],
                [2.5, 6],
                [1700, 2500],
                [0.025, 0.08]],
    'thetabounds': [[6, 10],
                    [-10, 10],
                    [16, 45],
                    [0.5, 1]],
    'c_array': np.arange(0.5, 1.5, 0.0005),
    'fail_array': np.array(
        [1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00,
         1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 1.000e+00, 9.998e-01,
         9.997e-01, 9.995e-01, 9.995e-01, 9.994e-01, 9.992e-01, 9.992e-01,
         9.991e-01, 9.989e-01, 9.987e-01, 9.984e-01, 9.982e-01, 9.978e-01,
         9.976e-01, 9.972e-01, 9.967e-01, 9.962e-01, 9.957e-01, 9.954e-01,
         9.947e-01, 9.942e-01, 9.935e-01, 9.929e-01, 9.923e-01, 9.916e-01,
         9.909e-01, 9.905e-01, 9.897e-01, 9.891e-01, 9.885e-01, 9.874e-01,
         9.862e-01, 9.856e-01, 9.848e-01, 9.839e-01, 9.833e-01, 9.825e-01,
         9.816e-01, 9.809e-01, 9.795e-01, 9.784e-01, 9.774e-01, 9.762e-01,
         9.751e-01, 9.742e-01, 9.731e-01, 9.720e-01, 9.707e-01, 9.693e-01,
         9.680e-01, 9.666e-01, 9.654e-01, 9.641e-01, 9.626e-01, 9.612e-01,
         9.597e-01, 9.582e-01, 9.563e-01, 9.546e-01, 9.529e-01, 9.516e-01,
         9.500e-01, 9.482e-01, 9.468e-01, 9.447e-01, 9.431e-01, 9.410e-01,
         9.394e-01, 9.375e-01, 9.358e-01, 9.341e-01, 9.324e-01, 9.304e-01,
         9.285e-01, 9.265e-01, 9.247e-01, 9.234e-01, 9.217e-01, 9.201e-01,
         9.182e-01, 9.165e-01, 9.148e-01, 9.134e-01, 9.120e-01, 9.103e-01,
         9.088e-01, 9.068e-01, 9.052e-01, 9.040e-01, 9.024e-01, 9.013e-01,
         8.997e-01, 8.982e-01, 8.966e-01, 8.954e-01, 8.942e-01, 8.927e-01,
         8.912e-01, 8.896e-01, 8.882e-01, 8.865e-01, 8.846e-01, 8.830e-01,
         8.813e-01, 8.800e-01, 8.784e-01, 8.766e-01, 8.752e-01, 8.739e-01,
         8.728e-01, 8.717e-01, 8.708e-01, 8.695e-01, 8.682e-01, 8.667e-01,
         8.652e-01, 8.636e-01, 8.623e-01, 8.604e-01, 8.584e-01, 8.566e-01,
         8.546e-01, 8.524e-01, 8.507e-01, 8.489e-01, 8.468e-01, 8.446e-01,
         8.423e-01, 8.403e-01, 8.385e-01, 8.364e-01, 8.343e-01, 8.325e-01,
         8.308e-01, 8.291e-01, 8.272e-01, 8.252e-01, 8.234e-01, 8.223e-01,
         8.207e-01, 8.192e-01, 8.179e-01, 8.167e-01, 8.158e-01, 8.146e-01,
         8.134e-01, 8.125e-01, 8.114e-01, 8.101e-01, 8.082e-01, 8.068e-01,
         8.053e-01, 8.035e-01, 8.020e-01, 8.004e-01, 7.991e-01, 7.976e-01,
         7.963e-01, 7.948e-01, 7.940e-01, 7.927e-01, 7.917e-01, 7.905e-01,
         7.892e-01, 7.882e-01, 7.870e-01, 7.858e-01, 7.844e-01, 7.829e-01,
         7.814e-01, 7.796e-01, 7.785e-01, 7.767e-01, 7.747e-01, 7.730e-01,
         7.721e-01, 7.706e-01, 7.690e-01, 7.675e-01, 7.663e-01, 7.651e-01,
         7.638e-01, 7.621e-01, 7.598e-01, 7.585e-01, 7.570e-01, 7.550e-01,
         7.531e-01, 7.512e-01, 7.495e-01, 7.479e-01, 7.460e-01, 7.438e-01,
         7.416e-01, 7.398e-01, 7.377e-01, 7.357e-01, 7.334e-01, 7.318e-01,
         7.302e-01, 7.286e-01, 7.271e-01, 7.257e-01, 7.241e-01, 7.220e-01,
         7.199e-01, 7.176e-01, 7.151e-01, 7.127e-01, 7.100e-01, 7.082e-01,
         7.064e-01, 7.045e-01, 7.029e-01, 7.014e-01, 6.998e-01, 6.988e-01,
         6.974e-01, 6.961e-01, 6.948e-01, 6.936e-01, 6.922e-01, 6.915e-01,
         6.909e-01, 6.895e-01, 6.889e-01, 6.872e-01, 6.859e-01, 6.843e-01,
         6.823e-01, 6.799e-01, 6.780e-01, 6.761e-01, 6.737e-01, 6.717e-01,
         6.696e-01, 6.679e-01, 6.666e-01, 6.650e-01, 6.637e-01, 6.627e-01,
         6.616e-01, 6.599e-01, 6.581e-01, 6.565e-01, 6.549e-01, 6.538e-01,
         6.529e-01, 6.517e-01, 6.501e-01, 6.483e-01, 6.459e-01, 6.443e-01,
         6.425e-01, 6.402e-01, 6.383e-01, 6.360e-01, 6.344e-01, 6.327e-01,
         6.313e-01, 6.295e-01, 6.274e-01, 6.257e-01, 6.238e-01, 6.220e-01,
         6.200e-01, 6.178e-01, 6.155e-01, 6.133e-01, 6.114e-01, 6.097e-01,
         6.087e-01, 6.075e-01, 6.062e-01, 6.049e-01, 6.034e-01, 6.016e-01,
         6.000e-01, 5.986e-01, 5.974e-01, 5.956e-01, 5.944e-01, 5.929e-01,
         5.904e-01, 5.884e-01, 5.868e-01, 5.857e-01, 5.845e-01, 5.826e-01,
         5.817e-01, 5.812e-01, 5.805e-01, 5.789e-01, 5.774e-01, 5.764e-01,
         5.749e-01, 5.726e-01, 5.703e-01, 5.680e-01, 5.652e-01, 5.628e-01,
         5.618e-01, 5.599e-01, 5.580e-01, 5.562e-01, 5.544e-01, 5.524e-01,
         5.502e-01, 5.480e-01, 5.459e-01, 5.443e-01, 5.425e-01, 5.403e-01,
         5.382e-01, 5.377e-01, 5.367e-01, 5.349e-01, 5.332e-01, 5.309e-01,
         5.298e-01, 5.276e-01, 5.263e-01, 5.240e-01, 5.219e-01, 5.190e-01,
         5.176e-01, 5.150e-01, 5.143e-01, 5.125e-01, 5.115e-01, 5.100e-01,
         5.074e-01, 5.050e-01, 5.020e-01, 4.995e-01, 4.980e-01, 4.964e-01,
         4.945e-01, 4.937e-01, 4.920e-01, 4.893e-01, 4.879e-01, 4.875e-01,
         4.866e-01, 4.855e-01, 4.828e-01, 4.804e-01, 4.775e-01, 4.747e-01,
         4.731e-01, 4.719e-01, 4.701e-01, 4.697e-01, 4.666e-01, 4.658e-01,
         4.643e-01, 4.629e-01, 4.620e-01, 4.590e-01, 4.558e-01, 4.544e-01,
         4.529e-01, 4.511e-01, 4.494e-01, 4.472e-01, 4.455e-01, 4.447e-01,
         4.433e-01, 4.417e-01, 4.404e-01, 4.384e-01, 4.365e-01, 4.340e-01,
         4.317e-01, 4.304e-01, 4.286e-01, 4.271e-01, 4.263e-01, 4.257e-01,
         4.250e-01, 4.238e-01, 4.228e-01, 4.216e-01, 4.202e-01, 4.183e-01,
         4.147e-01, 4.113e-01, 4.085e-01, 4.070e-01, 4.058e-01, 4.043e-01,
         4.028e-01, 4.009e-01, 3.995e-01, 3.981e-01, 3.966e-01, 3.949e-01,
         3.934e-01, 3.921e-01, 3.913e-01, 3.899e-01, 3.885e-01, 3.869e-01,
         3.844e-01, 3.827e-01, 3.801e-01, 3.780e-01, 3.761e-01, 3.739e-01,
         3.709e-01, 3.689e-01, 3.659e-01, 3.634e-01, 3.606e-01, 3.578e-01,
         3.556e-01, 3.538e-01, 3.519e-01, 3.501e-01, 3.483e-01, 3.468e-01,
         3.453e-01, 3.444e-01, 3.433e-01, 3.421e-01, 3.408e-01, 3.395e-01,
         3.384e-01, 3.371e-01, 3.349e-01, 3.328e-01, 3.312e-01, 3.292e-01,
         3.277e-01, 3.265e-01, 3.251e-01, 3.231e-01, 3.215e-01, 3.197e-01,
         3.183e-01, 3.168e-01, 3.151e-01, 3.137e-01, 3.119e-01, 3.102e-01,
         3.089e-01, 3.076e-01, 3.060e-01, 3.043e-01, 3.027e-01, 3.008e-01,
         2.987e-01, 2.970e-01, 2.954e-01, 2.942e-01, 2.927e-01, 2.910e-01,
         2.894e-01, 2.872e-01, 2.855e-01, 2.835e-01, 2.816e-01, 2.797e-01,
         2.779e-01, 2.759e-01, 2.741e-01, 2.722e-01, 2.710e-01, 2.690e-01,
         2.671e-01, 2.659e-01, 2.644e-01, 2.630e-01, 2.619e-01, 2.605e-01,
         2.594e-01, 2.586e-01, 2.576e-01, 2.560e-01, 2.544e-01, 2.529e-01,
         2.510e-01, 2.490e-01, 2.470e-01, 2.448e-01, 2.428e-01, 2.406e-01,
         2.381e-01, 2.356e-01, 2.332e-01, 2.311e-01, 2.292e-01, 2.270e-01,
         2.250e-01, 2.234e-01, 2.223e-01, 2.209e-01, 2.193e-01, 2.170e-01,
         2.156e-01, 2.142e-01, 2.127e-01, 2.113e-01, 2.093e-01, 2.076e-01,
         2.057e-01, 2.041e-01, 2.019e-01, 2.001e-01, 1.980e-01, 1.956e-01,
         1.929e-01, 1.903e-01, 1.876e-01, 1.853e-01, 1.831e-01, 1.810e-01,
         1.786e-01, 1.773e-01, 1.761e-01, 1.747e-01, 1.733e-01, 1.717e-01,
         1.702e-01, 1.682e-01, 1.662e-01, 1.648e-01, 1.636e-01, 1.616e-01,
         1.594e-01, 1.577e-01, 1.561e-01, 1.541e-01, 1.521e-01, 1.501e-01,
         1.485e-01, 1.465e-01, 1.447e-01, 1.427e-01, 1.408e-01, 1.394e-01,
         1.380e-01, 1.366e-01, 1.353e-01, 1.344e-01, 1.330e-01, 1.319e-01,
         1.303e-01, 1.287e-01, 1.270e-01, 1.250e-01, 1.230e-01, 1.209e-01,
         1.190e-01, 1.167e-01, 1.145e-01, 1.122e-01, 1.104e-01, 1.084e-01,
         1.069e-01, 1.054e-01, 1.037e-01, 1.024e-01, 1.001e-01, 9.807e-02,
         9.587e-02, 9.397e-02, 9.180e-02, 8.940e-02, 8.660e-02, 8.370e-02,
         8.063e-02, 7.763e-02, 7.527e-02, 7.307e-02, 7.063e-02, 6.860e-02,
         6.623e-02, 6.477e-02, 6.283e-02, 6.093e-02, 5.887e-02, 5.700e-02,
         5.553e-02, 5.397e-02, 5.280e-02, 5.073e-02, 4.893e-02, 4.710e-02,
         4.523e-02, 4.387e-02, 4.220e-02, 4.073e-02, 3.913e-02, 3.720e-02,
         3.583e-02, 3.420e-02, 3.260e-02, 3.160e-02, 3.047e-02, 2.967e-02,
         2.870e-02, 2.780e-02, 2.673e-02, 2.590e-02, 2.510e-02, 2.447e-02,
         2.377e-02, 2.313e-02, 2.223e-02, 2.123e-02, 2.047e-02, 1.940e-02,
         1.853e-02, 1.763e-02, 1.660e-02, 1.563e-02, 1.440e-02, 1.383e-02,
         1.333e-02, 1.267e-02, 1.170e-02, 1.070e-02, 9.833e-03, 8.933e-03,
         8.367e-03, 7.833e-03, 7.067e-03, 6.267e-03, 5.000e-03, 4.433e-03,
         3.700e-03, 3.167e-03, 2.800e-03, 2.400e-03, 2.200e-03, 2.000e-03,
         1.500e-03, 1.367e-03, 1.033e-03, 8.333e-04, 7.000e-04, 5.000e-04,
         3.333e-04, 3.000e-04, 2.333e-04, 1.667e-04, 1.333e-04, 6.667e-05,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00, 0.000e+00,
         0.000e+00, 0.000e+00])
}
